/* $Id: dbnj_hash_table.c,v 1.73 2005/11/16 03:44:00 oohara Exp $ */

#include <stdio.h>
/* malloc */
#include <stdlib.h>

#include "dbnj_util.h"
#include "dbnj_list.h"

#include "dbnj_hash_table.h"

static int dbnj_prime[] =
  {
    17,
    47,
    97,
    197,
    397,
    797,
    1597,
    3217,
    6427,
    12347,
    24697,
    50047,
    100057,
    200017,
    400067,
    800057,
    1600037,
    3200027,
    6400117
  };
static const int dbnj_prime_size = sizeof(dbnj_prime) / sizeof(dbnj_prime[0]);

static dbnj_hash_node *dbnj_hash_node_new(int n_errfile, FILE **p_errfile,
                                          void *key,
                                          void (*key_delete_func)(void *),
                                          void *data,
                                          void (*data_delete_func)(void *));
static void dbnj_hash_node_free(dbnj_hash_node *node);
static void dbnj_hash_node_delete(dbnj_hash_node *node);
static dbnj_list **dbnj_hash_p_node_new(int n_errfile, FILE **p_errfile,
                                             int size_node);
static dbnj_hash_table *dbnj_hash_table_alloc(int n_errfile, FILE **p_errfile);
static int dbnj_hash_table_sane(int n_errfile, FILE **p_errfile,
                                dbnj_hash_table *table);
static int dbnj_hash_table_lookup_list(int n_errfile, FILE **p_errfile,
                                       dbnj_hash_table *table,
                                       const void *target_key,
                                       dbnj_list **found_list);
static int dbnj_hash_table_resize(int n_errfile, FILE **p_errfile,
                                  dbnj_hash_table *table);

/* return NULL on error */
static dbnj_hash_node *
dbnj_hash_node_new(int n_errfile, FILE **p_errfile,
                   void *key,
                   void (*key_delete_func)(void *),
                   void *data,
                   void (*data_delete_func)(void *))
{
  dbnj_hash_node *result = NULL;

  result = (dbnj_hash_node *) malloc(sizeof(dbnj_hash_node));
  if (result == NULL)
  {
    dbnj_fprintf(n_errfile, p_errfile,
                 "dbnj_hash_node_new: malloc failed\n");
    return NULL;
  }

  result->key = key;
  result->key_delete_func = key_delete_func;
  result->data = data;
  result->data_delete_func = data_delete_func;

  return result;
}

/* this function does _NOT_ free key or data in node (arg 1) */
static void
dbnj_hash_node_free(dbnj_hash_node *node)
{
  if (node == NULL)
    return;

  node->key = NULL;
  node->key_delete_func = NULL;
  node->data = NULL;
  node->data_delete_func = NULL;
  free(node);
}

static void
dbnj_hash_node_delete(dbnj_hash_node *node)
{
  if (node == NULL)
    return;

  if ((node->key != NULL) && (node->key_delete_func != NULL))
    (*(node->key_delete_func))(node->key);
  if ((node->data != NULL) && (node->data_delete_func != NULL))
    (*(node->data_delete_func))(node->data);

  dbnj_hash_node_free(node);
}

/* return NULL on error */
static dbnj_list **
dbnj_hash_p_node_new(int n_errfile, FILE **p_errfile,
                     int size_node)
{
  int i;
  dbnj_list **result = NULL;

  /* sanity check */
  if (size_node <= 0)
  {
    dbnj_fprintf(n_errfile, p_errfile,
                 "dbnj_hash_p_node_new: size_node is non-positive (%d)\n",
                 size_node);
    return NULL;
  }

  result = (dbnj_list **) malloc(sizeof(dbnj_list *) * size_node);
  if (result == NULL)
  {
    dbnj_fprintf(n_errfile, p_errfile,
                 "dbnj_hash_p_node_new: malloc failed\n",
                 size_node);
    return NULL;
  }

  for (i = 0; i < size_node; i++)
    result[i] = NULL;

  return result;
}

/* return NULL on error */
static dbnj_hash_table *
dbnj_hash_table_alloc(int n_errfile, FILE **p_errfile)
{
  dbnj_hash_table *result = NULL;

  result = (dbnj_hash_table *) malloc(sizeof(dbnj_hash_table));
  if (result == NULL)
  {
    dbnj_fprintf(n_errfile, p_errfile,
                 "dbnj_hash_table_alloc: malloc failed\n");
    return NULL;
  }

  result->n_node = 0;
  result->size_node = 0;
  result->p_node = NULL;
  result->key_hash_func = NULL;
  result->key_equal_func = NULL;
  result->key_delete_func = NULL;

  return result;
}

/* return NULL on error */
dbnj_hash_table *
dbnj_hash_table_new(int n_errfile, FILE **p_errfile,
                    int (*key_hash_func)(const void *),
                    int (*key_equal_func)(const void *, const void *),
                    void (*key_delete_func)(void *))
{
  dbnj_hash_table *result = NULL;
  int size_node = 0;
  dbnj_list **p_node = NULL;

  /* sanity check */
  if (key_hash_func == NULL)
  {
    dbnj_fprintf(n_errfile, p_errfile,
                 "dbnj_hash_table_new: key_hash_func is NULL\n");
    return NULL;
  }
  if (key_equal_func == NULL)
  {
    dbnj_fprintf(n_errfile, p_errfile,
                 "dbnj_hash_table_new: key_equal_func is NULL\n");
    return NULL;
  }

  result = dbnj_hash_table_alloc(n_errfile, p_errfile);
  if (result == NULL)
  {
    dbnj_fprintf(n_errfile, p_errfile,
                 "dbnj_hash_table_new: dbnj_hash_table_alloc failed\n");
    return NULL;
  }

  size_node = dbnj_prime[0];
  p_node = dbnj_hash_p_node_new(n_errfile, p_errfile, size_node);
  if (p_node == NULL)
  {
    dbnj_fprintf(n_errfile, p_errfile,
                 "dbnj_hash_table_new: dbnj_hash_p_node_new failed\n");
    dbnj_hash_table_delete(n_errfile, p_errfile,
                           result);
    return NULL;
  }

  result->size_node = size_node;
  result->p_node = p_node;
  result->key_hash_func = key_hash_func;
  result->key_equal_func = key_equal_func;
  result->key_delete_func = key_delete_func;

  return result;
}

/* return 1 (true) or 0 (false) */
static int
dbnj_hash_table_sane(int n_errfile, FILE **p_errfile,
                     dbnj_hash_table *table)
{
  if (table == NULL)
  {
    dbnj_fprintf(n_errfile, p_errfile,
                 "dbnj_hash_table_sane: table is NULL\n");
    return 0;
  }
  if (table->n_node < 0)
  {
    dbnj_fprintf(n_errfile, p_errfile,
                 "dbnj_hash_table_sane: table->n_node is negative "
                 "(%d)\n", table->n_node);
    return 0;
  }
  if (table->size_node <= 0)
  {
    dbnj_fprintf(n_errfile, p_errfile,
                 "dbnj_hash_table_sane: table->size_node is non-positive "
                 "(%d)\n", table->size_node);
    return 0;
  }
  if (table->p_node == NULL)
  {
    dbnj_fprintf(n_errfile, p_errfile,
                 "dbnj_hash_table_sane: table->p_node is NULL\n");
    return 0;
  }
  if (table->key_hash_func == NULL)
  {
    dbnj_fprintf(n_errfile, p_errfile,
                 "dbnj_hash_table_sane: table->key_hash_func is NULL\n");
    return 0;
  }
  if (table->key_equal_func == NULL)
  {
    dbnj_fprintf(n_errfile, p_errfile,
                 "dbnj_hash_table_sane: table->key_equal_func is NULL\n");
    return 0;
  }

  return 1;
}

void
dbnj_hash_table_delete(int n_errfile, FILE **p_errfile,
                       dbnj_hash_table *table)
{
  int i;

  if (table == NULL)
    return;

  if (table->p_node != NULL)
  {
    for (i = 0; i < table->size_node; i++)
    {
      if (table->p_node[i] != NULL)
      {
        dbnj_list_delete(n_errfile, p_errfile,
                         table->p_node[i]);
        table->p_node[i] = NULL;
      }
    }
    free(table->p_node);
    table->p_node = NULL;
  }

  table->n_node = 0;
  table->size_node = 0;
  table->p_node = NULL;
  table->key_hash_func = NULL;
  table->key_equal_func = NULL;
  table->key_delete_func = NULL;

  free(table);
}

/* if replace (arg 4) is non-zero, the key must be in the table
 * if replace (arg 4) is zero, the key must not be in the table
 * return 0 on success, 1 on error
 */
int
dbnj_hash_table_insert(int n_errfile, FILE **p_errfile,
                       dbnj_hash_table *table, int replace,
                       void *key, void *data,
                       void (*data_delete_func)(void *))
{
  int n;
  int found;
  dbnj_list *found_list = NULL;
  dbnj_hash_node *node = NULL;
  dbnj_list *new_list = NULL;

  /* sanity check */
  if (!dbnj_hash_table_sane(n_errfile, p_errfile,
                            table))
  {
    dbnj_fprintf(n_errfile, p_errfile,
                 "dbnj_hash_table_insert: table is not sane\n");
    return 1;
  }
  if (key == NULL)
  {
    dbnj_fprintf(n_errfile, p_errfile,
                 "dbnj_hash_table_insert: key is NULL\n");
    return 1;
  }
  /* data == NULL is OK */

  found_list = NULL;
  found = dbnj_hash_table_lookup_list(n_errfile, p_errfile,
                                      table, key,
                                      &found_list);
  if (replace)
  {
    if (!found)
    {
      dbnj_fprintf(n_errfile, p_errfile,
                   "dbnj_hash_table_insert: key not found when replacing\n");
      return 1;
    }
    if (found_list == NULL)
    {
      dbnj_fprintf(n_errfile, p_errfile,
                   "dbnj_hash_table_insert: dbnj_hash_table_lookup_list set "
                   "found_list to NULL while returning non-zero\n");
      return 1;
    }
    node = (dbnj_hash_node *) (found_list->data);

    if ((node->key != NULL) && (node->key_delete_func != NULL))
      (*(node->key_delete_func))(node->key);
    if ((node->data != NULL) && (node->data_delete_func != NULL))
      (*(node->data_delete_func))(node->data);

    node->key = key;
    /* no need to touch node->key_delete_func */
    node->data = data;
    node->data_delete_func = data_delete_func;
    return 0;
  }
  else
  {
    if (found)
    {
      dbnj_fprintf(n_errfile, p_errfile,
                   "dbnj_hash_table_insert: key already in the table "
                   "when inserting\n");
      return 1;
    }

    n = (*(table->key_hash_func))(key);
    if (n < 0)
    {
      dbnj_fprintf(n_errfile, p_errfile,
                   "dbnj_hash_table_insert: hash returned a negative value "
                   "(%d)\n", n);
      return 1;
    }

    n = n % table->size_node;
    found_list = table->p_node[n];

    node = dbnj_hash_node_new(n_errfile, p_errfile,
                              key,
                              table->key_delete_func,
                              data,
                              data_delete_func);
    if (node == NULL)
    {
      dbnj_fprintf(n_errfile, p_errfile,
                   "dbnj_hash_table_insert: dbnj_hash_node_new failed\n");
      return 1;
    }

    new_list = dbnj_list_new(n_errfile, p_errfile,
                             (void *) node,
                             (void (*)(void *)) dbnj_hash_node_delete);
    if (new_list == NULL)
    {
      dbnj_fprintf(n_errfile, p_errfile,
                   "dbnj_hash_table_insert: dbnj_list_new failed\n");
      /* don't use dbnj_hash_node_delete here because key or data is not
       * in the table yet
      */
      dbnj_hash_node_free(node);
      return 1;
    }

    table->p_node[n] = dbnj_list_concat(n_errfile, p_errfile,
                                        found_list, new_list);
    if (table->p_node[n] == NULL)
    {
      dbnj_fprintf(n_errfile, p_errfile,
                   "dbnj_hash_table_insert: dbnj_list_concat failed\n");
      /* I don't know how to unwind here */
      return 1;
    }

    (table->n_node)++;
    dbnj_hash_table_resize(n_errfile, p_errfile,
                           table);
    return 0;
  }

  return 0;
}

/* if found_list (arg 5) not NULL,
 * set to the key-data pair whose key is same as target_key (arg 4),
 * or NULL it the key is not in the table
 * return 1 if the key is in the table, 0 otherwise
 */
static int
dbnj_hash_table_lookup_list(int n_errfile, FILE **p_errfile,
                            dbnj_hash_table *table,
                            const void *target_key,
                            dbnj_list **found_list)
{
  int n;
  dbnj_list *list = NULL;
  dbnj_hash_node *node = NULL;

  if (found_list != NULL)
  {
    *found_list = NULL;
  }

  /* sanity check */
  if (!dbnj_hash_table_sane(n_errfile, p_errfile,
                            table))
  {
    dbnj_fprintf(n_errfile, p_errfile,
                 "dbnj_hash_table_lookup_list: table is not sane\n");
    return 0;
  }
  if (target_key == NULL)
  {
    dbnj_fprintf(n_errfile, p_errfile,
                 "dbnj_hash_table_lookup_list: target_key is NULL\n");
    return 0;
  }

  n = (*(table->key_hash_func))(target_key);
  if (n < 0)
  {
    dbnj_fprintf(n_errfile, p_errfile,
                 "dbnj_hash_table_lookup_list: hash returned a negative value "
                 "(%d)\n", n);
    return 0;
  }

  n = n % table->size_node;
  list = table->p_node[n];
  if (list == NULL)
  {
    return 0;
  }
  if (list->prev != NULL)
  {
    dbnj_fprintf(n_errfile, p_errfile,
                 "dbnj_hash_table_lookup_list: table->p_node[%d] is not "
                 "the head of a list\n", n);
    return 0;
  }
  while (list != NULL)
  {
    node = (dbnj_hash_node *) (list->data);
    if ((*(table->key_equal_func))(target_key, node->key))
    {
      if (found_list != NULL)
      {
        /* not target_key */
        *found_list = list;
      }

      return 1;
    }
    list = list->next;
  }

  return 0;
}

/* if found_key (arg 5) and/or found_data (arg 6) are not NULL,
 * set to the key-data pair whose key is same as target_key (arg 4),
 * or NULL it the key is not in the table
 * return 1 if the key is in the table, 0 otherwise
 */
int
dbnj_hash_table_lookup(int n_errfile, FILE **p_errfile,
                       dbnj_hash_table *table,
                       const void *target_key,
                       void **found_key, void **found_data)
{
  int found;
  dbnj_list *found_list = NULL;
  dbnj_hash_node *node = NULL;

  if (found_key != NULL)
  {
    *found_key = NULL;
  }
  if (found_data != NULL)
  {
    *found_data = NULL;
  }

  /* sanity check */
  if (!dbnj_hash_table_sane(n_errfile, p_errfile,
                            table))
  {
    dbnj_fprintf(n_errfile, p_errfile,
                 "dbnj_hash_table_lookup: table is not sane\n");
    return 0;
  }
  if (target_key == NULL)
  {
    dbnj_fprintf(n_errfile, p_errfile,
                 "dbnj_hash_table_lookup: target_key is NULL\n");
    return 0;
  }

  found_list = NULL;
  found = dbnj_hash_table_lookup_list(n_errfile, p_errfile,
                                      table, target_key,
                                      &found_list);
  if (!found)
    return 0;
  if (found_list == NULL)
  {
    dbnj_fprintf(n_errfile, p_errfile,
                 "dbnj_hash_table_lookup: dbnj_hash_table_lookup_list "
                 "set found_list to NULL while returning non-zero\n");
    return 0;
  }

  node = (dbnj_hash_node *) (found_list->data);
  if (found_key != NULL)
  {
    /* not target_key */
    *found_key = node->key;
  }
  if (found_data != NULL)
  {
    *found_data = node->data;
  }

  return 1;
}

/* target_key (arg 4) is not deleted by this function
 * return 0 on success, 1 on error
 */
int
dbnj_hash_table_remove(int n_errfile, FILE **p_errfile,
                       dbnj_hash_table *table,
                       const void *target_key)
{
  int n;
  int found;
  dbnj_list *found_list = NULL;
  dbnj_list *new_list = NULL;

  /* sanity check */
  if (!dbnj_hash_table_sane(n_errfile, p_errfile,
                            table))
  {
    dbnj_fprintf(n_errfile, p_errfile,
                 "dbnj_hash_table_remove: table is not sane\n");
    return 1;
  }
  if (target_key == NULL)
  {
    dbnj_fprintf(n_errfile, p_errfile,
                 "dbnj_hash_table_remove: target_key is NULL\n");
    return 1;
  }

  found_list = NULL;
  found = dbnj_hash_table_lookup_list(n_errfile, p_errfile,
                                      table, target_key,
                                      &found_list);
  if (!found)
  {
    dbnj_fprintf(n_errfile, p_errfile,
                 "dbnj_hash_table_remove: target_key is not in the table\n");
    return 1;
  }
  if (found_list == NULL)
  {
    dbnj_fprintf(n_errfile, p_errfile,
                 "dbnj_hash_table_remove: dbnj_hash_table_lookup_list "
                 "set found_list to NULL while returning non-zero\n");
    return 1;
  }

  new_list = dbnj_list_remove_link(n_errfile, p_errfile,
                                   found_list, found_list);
  if (dbnj_list_delete(n_errfile, p_errfile,
                       found_list) != 0)
  {
    dbnj_fprintf(n_errfile, p_errfile,
                 "dbnj_hash_table_remove: dbnj_list_delete failed\n");
    return 1;
  }

  n = (*(table->key_hash_func))(target_key);
  if (n < 0)
  {
    dbnj_fprintf(n_errfile, p_errfile,
                 "dbnj_hash_table_remove: hash returned a negative value "
                 "(%d)\n", n);
    return 1;
  }

  n = n % table->size_node;
  table->p_node[n] = dbnj_list_first(n_errfile, p_errfile,
                                     new_list);
  (table->n_node)--;
  dbnj_hash_table_resize(n_errfile, p_errfile,
                         table);

  return 0;
}

/* return 0 on success, 1 on error
 * assumes that dbnj_prime is sorted
 */
static int
dbnj_hash_table_resize(int n_errfile, FILE **p_errfile,
                       dbnj_hash_table *table)
{
  int i;
  int n;
  int size_node_new;
  dbnj_list **p_node_new = NULL;
  dbnj_list *list_temp = NULL;
  dbnj_list *list = NULL;
  dbnj_list *p = NULL;

  /* sanity check */
  if (table == NULL)
  {
    dbnj_fprintf(n_errfile, p_errfile,
                 "dbnj_hash_table_resize: table is NULL\n");
    return 1;
  }
  if (dbnj_prime_size <= 0)
  {
    dbnj_fprintf(n_errfile, p_errfile,
                 "dbnj_hash_table_resize: dbnj_prime_size is non-positive "
                 "(%d)\n", dbnj_prime_size);
    return 1;
  }

  if ((table->n_node * 2 <= table->size_node)
      && (table->n_node * 5 >= table->size_node))
    return 0;

  size_node_new = dbnj_prime[dbnj_prime_size - 1];
  for (i = 0; i < dbnj_prime_size; i++)
  {
    if (dbnj_prime[i] >= table->n_node * 2)
    {
      size_node_new = dbnj_prime[i];
      break;
    }
  }

  if (size_node_new == table->size_node)
    return 0;

  list_temp = NULL;  
  for (i = 0; i < table->size_node; i++)
  {
    list_temp = dbnj_list_concat(n_errfile, p_errfile,
                                 list_temp, table->p_node[i]);
    table->p_node[i] = NULL;
  }

  p_node_new = (dbnj_list **) realloc(table->p_node,
                                      sizeof(dbnj_list *) * size_node_new);
  if (p_node_new == NULL)
  {
    dbnj_fprintf(n_errfile, p_errfile,
                 "dbnj_hash_table_resize: realloc failed\n");
    size_node_new = table->size_node;
  }
  else
  {
    table->size_node = size_node_new;
    table->p_node = p_node_new;
    for (i = 0; i < table->size_node; i++)
    {
      table->p_node[i] = NULL;
    }
  }

  while (list_temp != NULL)
  {
    p = list_temp;
    list_temp = dbnj_list_remove_link(n_errfile, p_errfile,
                                      p, p);

    n = (*(table->key_hash_func))(((dbnj_hash_node *) (p->data))->key);
    if (n < 0)
    {
      dbnj_fprintf(n_errfile, p_errfile,
                   "dbnj_hash_table_resize: hash returned a negative value "
                   "(%d)\n", n);
      n = -n;
    }

    n = n % table->size_node;
    list = table->p_node[n];
    table->p_node[n] = dbnj_list_concat(n_errfile, p_errfile,
                                        list, p);
  }

  return 0;
}
