/* $Id: dbnj_list.c,v 1.44 2005/10/24 19:57:56 oohara Exp $ */

#include <stdio.h>
/* malloc */
#include <stdlib.h>

#include "dbnj_util.h"

#include "dbnj_list.h"

static dbnj_list *dbnj_list_alloc(int n_errfile, FILE **p_errfile);
static void dbnj_list_free(dbnj_list *list);
static int dbnj_list_delete_1(int n_errfile, FILE **p_errfile,
                              dbnj_list *list);
static int dbnj_list_has_this_link(int n_errfile, FILE **p_errfile,
                                   dbnj_list *list, dbnj_list *target);

/* NULL is an empty dbnj_list */

/* return a list that has NULL as the only data */
static dbnj_list *
dbnj_list_alloc(int n_errfile, FILE **p_errfile)
{
  dbnj_list *new_list = NULL;

  new_list = (dbnj_list *) malloc(sizeof(dbnj_list));
  if (new_list == NULL)
  {
    dbnj_fprintf(n_errfile, p_errfile,
                 "dbnj_list_alloc: malloc failed\n");
    return NULL;
  }

  new_list->prev = NULL;
  new_list->next = NULL;
  new_list->data = NULL;
  new_list->data_delete_func = NULL;
  return new_list;
}

/* this function does _NOT_ free the data in the list (arg 3) */
static void
dbnj_list_free(dbnj_list *list)
{
  if (list == NULL)
    return;

  list->prev = NULL;
  list->next = NULL;
  list->data = NULL;
  list->data_delete_func = NULL;

  free(list);
}

/* return a list of length 1 whose only data is data (arg 3)
 * return NULL on error
 */
dbnj_list *
dbnj_list_new(int n_errfile, FILE **p_errfile,
              void *data, void (*data_delete_func)(void *))
{
  dbnj_list *new_list = NULL;

  new_list = dbnj_list_alloc(n_errfile, p_errfile);
  if (new_list == NULL)
  {
    dbnj_fprintf(n_errfile, p_errfile,
                 "dbnj_list_new: dbnj_list_alloc failed\n");
    return NULL;
  }
  new_list->data = data;
  new_list->data_delete_func = data_delete_func;

  return new_list;
}

/* the length of list (arg 3) must be 1
 * return 0 on success, 1 on error
 */
static int
dbnj_list_delete_1(int n_errfile, FILE **p_errfile,
                   dbnj_list *list)
{
  /* sanity check */
  if (list == NULL)
  {
    dbnj_fprintf(n_errfile, p_errfile,
                 "dbnj_list_delete_1: list is NULL\n");
    return 1;
  }
  if (list->prev != NULL)
  {
    dbnj_fprintf(n_errfile, p_errfile,
                 "dbnj_list_delete_1: list->prev is not NULL\n");
    return 1;
  }
  if (list->next != NULL)
  {
    dbnj_fprintf(n_errfile, p_errfile,
                 "dbnj_list_delete_1: list->next is not NULL\n");
    return 1;
  }

  if ((list->data != NULL) && (list->data_delete_func != NULL))
    (*(list->data_delete_func))(list->data);
  list->data = NULL;
  list->data_delete_func = NULL;
  dbnj_list_free(list);

  return 0;
}

/* delete the entire list
 * return 0 on success, 1 on error
 */
int
dbnj_list_delete(int n_errfile, FILE **p_errfile,
                 dbnj_list *list)
{
  dbnj_list *p = NULL;
  dbnj_list *q = NULL;

  p = list;
  while(p != NULL)
  {
    q = dbnj_list_remove_link(n_errfile, p_errfile,
                              p, p);
    if (dbnj_list_delete_1(n_errfile, p_errfile,
                           p) != 0)
    {
      dbnj_fprintf(n_errfile, p_errfile,
                 "dbnj_list_delete: dbnj_list_delete_1 failed\n");
      return 1;
    }
    p = q;
  }

  return 0;
}

/* return NULL on error */
dbnj_list *
dbnj_list_first(int n_errfile, FILE **p_errfile,
                dbnj_list *list)
{
  dbnj_list *p = NULL;

  if (list == NULL)
    return NULL;

  p = list;
  while (p->prev != NULL)
  {
    if (p->prev == list)
    {
      dbnj_fprintf(n_errfile, p_errfile,
                   "dbnj_list_first: circluar list found\n");
      return NULL;
    }
    p = p->prev;
  }

  return p;
}

/* return NULL on error */
dbnj_list *
dbnj_list_last(int n_errfile, FILE **p_errfile,
               dbnj_list *list)
{
  dbnj_list *p = NULL;

  if (list == NULL)
    return NULL;

  p = list;
  while (p->next != NULL)
  {
    if (p->next == list)
    {
      dbnj_fprintf(n_errfile, p_errfile,
                   "dbnj_list_last: circluar list found\n");
      return NULL;
    }
    p = p->next;
  }

  return p;
}

dbnj_list *
dbnj_list_nth(int n_errfile, FILE **p_errfile,
              dbnj_list *list, int n)
{
  int i;
  dbnj_list *p = NULL;

  if (list == NULL)
    return NULL;
  if (n == 0)
    return NULL;

  p = list;
  if (n > 0)
  {  
    for (i = 0; (i < n) && (p != NULL); i++)
      p = p->next;
  }
  else
  {  
    for (i = 0; (i > n) && (p != NULL); i--)
      p = p->prev;
  }

  return p;
}

/* return -1 on error */
int
dbnj_list_length_before(int n_errfile, FILE **p_errfile,
                        dbnj_list *list)
{
  int n;
  dbnj_list *p = NULL;

  if (list == NULL)
    return 0;

  p = list;
  n = 1;
  while (p->next != NULL)
  {
    if (p->next == list)
    {
      dbnj_fprintf(n_errfile, p_errfile,
                   "dbnj_list_length_before: circluar list found\n");
      return -1;
    }
    p = p->next;
    n++;
  }

  return n;
}

/* return -1 on error */
int
dbnj_list_length_after(int n_errfile, FILE **p_errfile,
                       dbnj_list *list)
{
  int n;
  dbnj_list *p = NULL;

  if (list == NULL)
    return 0;

  p = list;
  n = 1;
  while (p->next != NULL)
  {
    if (p->next == list)
    {
      dbnj_fprintf(n_errfile, p_errfile,
                   "dbnj_list_length_after: circluar list found\n");
      return -1;
    }
    p = p->next;
    n++;
  }

  return n;
}

/* return -1 on error */
int
dbnj_list_length_total(int n_errfile, FILE **p_errfile,
                       dbnj_list *list)
{
  int n_before;
  int n_after;
  int result;

  /* this check is necessary because list is not counted twice in this case */
  if (list == NULL)
    return 0;

  n_before = dbnj_list_length_before(n_errfile, p_errfile,
                                     list);
  if (n_before < 0)
  {
    dbnj_fprintf(n_errfile, p_errfile,
                 "dbnj_list_total: dbnj_list_length_before failed\n");
    return -1;
  }

  n_after = dbnj_list_length_after(n_errfile, p_errfile,
                                   list);
  if (n_before < 0)
  {
    dbnj_fprintf(n_errfile, p_errfile,
                 "dbnj_list_total: dbnj_list_length_after failed\n");
    return -1;
  }

  result = n_before + n_after - 1;
  if (result < 0)
  {
    dbnj_fprintf(n_errfile, p_errfile,
                 "dbnj_list_total: strange result (%d)\n", result);
    return -1;
  }

  return result;
}

/* return a new list on succrss, NULL on error */
dbnj_list *
dbnj_list_concat(int n_errfile, FILE **p_errfile,
                 dbnj_list *list1, dbnj_list *list2)
{
  dbnj_list *list1_last = NULL;
  dbnj_list *list2_first = NULL;
  dbnj_list *result = NULL;

  if (list1 == NULL)
    return list2;
  if (list2 == NULL)
    return list1;

  list1_last = dbnj_list_last(n_errfile, p_errfile,
                              list1);
  if (list1_last == NULL)
  {
    dbnj_fprintf(n_errfile, p_errfile,
                 "dbnj_list_concat: dbnj_list_last failed\n");
    return NULL;
  }
  list2_first = dbnj_list_first(n_errfile, p_errfile,
                                list2);
  if (list2_first == NULL)
  {
    dbnj_fprintf(n_errfile, p_errfile,
                 "dbnj_list_concat: dbnj_list_first failed\n");
    return NULL;
  }

  result = list1;
  list1_last->next = list2_first;
  list2_first->prev = list1_last;

  return result;
}

/* return 1 (true) or 0 (false) */
static int
dbnj_list_has_this_link(int n_errfile, FILE **p_errfile,
                        dbnj_list *list, dbnj_list *target)
{
  dbnj_list *p = NULL;

  /* sanity check */
  if (target == NULL)
  {
    dbnj_fprintf(n_errfile, p_errfile,
                 "dbnj_list_has_this_link: target is NULL\n");
    return 0;
  }

  if (list == NULL)
  {
    /* empty list */
    return 0;
  }
  if (list == target)
    return 1;

  p = list;
  while (p->next != NULL)
  {
    if (p->next == list)
    {
      dbnj_fprintf(n_errfile, p_errfile,
                   "dbnj_list_has_this_link: circluar list found "
                   "when searching forward\n");
      return 0;
    }

    if (p->next == target)
      return 1;

    p = p->next;
  }

  p = list;
  while (p->prev != NULL)
  {
    if (p->prev == list)
    {
      dbnj_fprintf(n_errfile, p_errfile,
                   "dbnj_list_has_this_link: circluar list found "
                   "when searching backward\n");
      return 0;
    }

    if (p->prev == target)
      return 1;

    p = p->prev;
  }

  return 0;
}

/* remove target (arg 4) from list (arg 3) and make target (arg 4)
 * a list of length 1 (use dbnj_list_delete to free its memory)
 * target (arg 4) must be in list (arg 3)
 * return a new list (which may be NULL)
 */
dbnj_list *
dbnj_list_remove_link(int n_errfile, FILE **p_errfile,
                      dbnj_list *list, dbnj_list *target)
{
  dbnj_list *result = NULL;

  /* sanity check */
  if (list == NULL)
  {
    dbnj_fprintf(n_errfile, p_errfile,
                 "dbnj_list_remove_link: list is NULL\n");
    return NULL;
  }
  if (target == NULL)
  {
    dbnj_fprintf(n_errfile, p_errfile,
                 "dbnj_list_remove_link: target is NULL\n");
    return list;
  }
  if (!dbnj_list_has_this_link(n_errfile, p_errfile,
                               list, target))
  {
    dbnj_fprintf(n_errfile, p_errfile,
                 "dbnj_list_remove_link: target is not in list\n");
    return list;
  }

  result = list;
  if (list == target)
  {
    if (list->next != NULL)
      result = list->next;
    else
      result = list->prev;
  }

  if (target->prev != NULL)
    target->prev->next = target->next;
  if (target->next != NULL)
    target->next->prev = target->prev;
  target->prev = NULL;
  target->next = NULL;

  return result;
}
