/* $Id: dbnj_sdl.c,v 1.38 2007/05/17 02:00:54 oohara Exp $ */

#include <stdio.h>
/* atexit */
#include <stdlib.h>

#include <SDL.h>
/* libsdl-ttf2.0 */
#include <SDL_ttf.h>

#include "dbnj_util.h"

#include "dbnj_sdl.h"

static Uint32 dbnj_map_rgb(int n_errfile, FILE **p_errfile,
                           SDL_Surface *surface,
                           int red, int green, int blue);

/* return a pointer to the current display on success
 * return NULL on error
 */
SDL_Surface *
dbnj_sdl_init(int n_errfile, FILE **p_errfile,
              int x, int y, int full_screen,
              const char *name)
{
  Uint32 flags;
  SDL_Surface *screen = NULL;

  /* sanity check */
  if (x <= 0)
  {
    dbnj_fprintf(n_errfile, p_errfile,
                 "dbnj_sdl_init: x is non-positive (%d)\n", x);
    return NULL;
  }
  if (y <= 0)
  {
    dbnj_fprintf(n_errfile, p_errfile,
                 "dbnj_sdl_init: y is non-positive (%d)\n", y);
    return NULL;
  }

  flags = 0;
  if (full_screen)
    flags |= SDL_FULLSCREEN;

  if (SDL_Init(SDL_INIT_VIDEO) != 0)
  {
    dbnj_fprintf(n_errfile, p_errfile,
                 "dbnj_sdl_init: cannot init SDL video");
    if (SDL_GetError() != NULL)
      dbnj_fprintf(n_errfile, p_errfile,
                   " (%s)", SDL_GetError());
    dbnj_fprintf(n_errfile, p_errfile,
                 "\n");
    return NULL;
  }
  if (atexit(SDL_Quit) != 0)
  {
    dbnj_fprintf(n_errfile, p_errfile,
                 "dbnj_sdl_init: atexit(SDL_Quit) failed\n");
    SDL_Quit();
    return NULL;
  }
  if (SDL_InitSubSystem(SDL_INIT_TIMER) != 0)
  {
    dbnj_fprintf(n_errfile, p_errfile,
                 "dbnj_sdl_init: cannot init SDL timer");
    if (SDL_GetError() != NULL)
      dbnj_fprintf(n_errfile, p_errfile,
                   " (%s)", SDL_GetError());
    dbnj_fprintf(n_errfile, p_errfile,
                 "\n");
    return NULL;
  }

  if (TTF_Init() < 0)
  {
    dbnj_fprintf(n_errfile, p_errfile,
                 "dbnj_sdl_init: cannot init SDL TTF");
    if (SDL_GetError() != NULL)
      dbnj_fprintf(n_errfile, p_errfile,
                   " (%s)", SDL_GetError());
    dbnj_fprintf(n_errfile, p_errfile,
                 "\n");
    return NULL;
  }
  if (atexit(TTF_Quit) != 0)
  {
    dbnj_fprintf(n_errfile, p_errfile,
                 "dbnj_sdl_init: atexit(TTF_Quit) failed\n");
    TTF_Quit();
    return NULL;
  }

  screen = SDL_SetVideoMode(x, y, 0, flags);
  if (screen == NULL)
  {
    dbnj_fprintf(n_errfile, p_errfile,
                 "dbnj_sdl_init: cannot set SDL video mode");
    if (SDL_GetError() != NULL)
      dbnj_fprintf(n_errfile, p_errfile,
                   " (%s)", SDL_GetError());
    dbnj_fprintf(n_errfile, p_errfile,
                 "\n");
    return NULL;
  }

  SDL_ShowCursor(SDL_DISABLE);
  if (name != NULL)
  SDL_WM_SetCaption(name, name);
  SDL_EnableUNICODE(1);

  return screen;
}

int
dbnj_sdl_clock_granularity(void)
{
  int tick_begin;
  int t;
  int i;
  int loop;

  loop = 10;
  if (loop < 1)
    loop = 1;

  /* jump to another 10 ms */
  SDL_Delay(10);
  tick_begin = SDL_GetTicks();
  for (i = 0; i < loop; i++)
    SDL_Delay(1);
  t = SDL_GetTicks() - tick_begin;
  if (t > 0)
    t = t / loop + 1;

  if (t < 1)
    t = 1;

  return t;
}

void
dbnj_sdl_wait(int tick_begin, int ms, int clock_granularity)
{
  /* sanity check */
  if (ms <= 0)
    return;

  if (clock_granularity > 0)
    while (SDL_GetTicks() + clock_granularity < tick_begin + ms)
      SDL_Delay(1);
  while (SDL_GetTicks() < tick_begin + ms)
    /* empty */;
}

/* return NULL on error */
TTF_Font *
dbnj_open_font(int n_errfile, FILE **p_errfile,
               const char *ttf_font_file, int ptsize)
{
  TTF_Font *font = NULL;

  /* sanity check */
  if (ttf_font_file == NULL)
  {
    dbnj_fprintf(n_errfile, p_errfile,
                 "dbnj_open_font: ttf_font_file is NULL\n");
    return NULL;
  }
  if (ptsize <= 0)
  {
    dbnj_fprintf(n_errfile, p_errfile,
                 "dbnj_open_font: ptsize is non-positive (%d)\n", ptsize);
    return NULL;
  }
  
  font = TTF_OpenFont(ttf_font_file, ptsize);
  if (font == NULL)
  {
    dbnj_fprintf(n_errfile, p_errfile,
                 "dbnj_open_font: cannot open font %s at size %d",
                 ttf_font_file, ptsize);
    if (SDL_GetError() != NULL)
      dbnj_fprintf(n_errfile, p_errfile,
                   " (%s)", SDL_GetError());
    dbnj_fprintf(n_errfile, p_errfile,
                 "\n");
    return NULL;
  }

  return font;
}

static Uint32
dbnj_map_rgb(int n_errfile, FILE **p_errfile,
             SDL_Surface *surface,
             int red, int green, int blue)
{
  /* sanity check */
  if (surface == NULL)
  {
    dbnj_fprintf(n_errfile, p_errfile,
                 "dbnj_map_rgb: surface is NULL");
    return 0;
  }
  if ((red < 0) || (red > 255))
  {
    dbnj_fprintf(n_errfile, p_errfile,
                 "dbnj_map_rgb: red (%d) is not in [0, 255]\n", red);
    return 0;
  }
  if ((green < 0) || (green > 255))
  {
    dbnj_fprintf(n_errfile, p_errfile,
                 "dbnj_map_rgb: green (%d) is not in [0, 255]\n", green);
    return 0;
  }
  if ((blue < 0) || (blue > 255))
  {
    dbnj_fprintf(n_errfile, p_errfile,
                 "dbnj_map_rgb: blue (%d) is not in [0, 255]\n", blue);
    return 0;
  }

  return SDL_MapRGB(surface->format, red, green, blue);
}

/* return 0 on success, 1 on error */
int
dbnj_draw_rectangle(int n_errfile, FILE **p_errfile,
                    SDL_Surface *surface,
                    int x, int y, int w, int h,
                    int red, int green, int blue)
{
  SDL_Rect rect;
  Uint32 color;

  /* sanity check */
  if (surface == NULL)
  {
    dbnj_fprintf(n_errfile, p_errfile,
                 "dbnj_draw_rectangle: surface is NULL\n");
    return 1;
  }
  if (w <= 0)
  {
    dbnj_fprintf(n_errfile, p_errfile,
                 "dbnj_draw_rectangle: w is non-positive (%d)\n", w);
    return 1;
  }
  if (h <= 0)
  {
    dbnj_fprintf(n_errfile, p_errfile,
                 "dbnj_draw_rectangle: h is non-positive (%d)\n", h);
    return 1;
  }

  rect.x = x;
  rect.y = y;
  rect.w = w;
  rect.h = h;
  color = dbnj_map_rgb(n_errfile, p_errfile,
                       surface,
                       red, green, blue);

  if (SDL_FillRect(surface, &rect, color) != 0)
  {
    dbnj_fprintf(n_errfile, p_errfile,
                 "dbnj_draw_rectangle: SDL_FillRect failed");
    if (SDL_GetError() != NULL)
      dbnj_fprintf(n_errfile, p_errfile,
                   " (%s)", SDL_GetError());
    dbnj_fprintf(n_errfile, p_errfile,
                 "\n");
    return 1;
  }

  return 0;
}

/* aligned (arg 8) is one of:
 *   0: left-aligned ((x, y) is the upper-left corner)
 *   1: right-aligned ((x, y) is the upper-right corner)
 *   2: center-aligned (x is the center, y is the upper side)
 * return 0 on success, 1 on error
 */
int dbnj_draw_text(int n_errfile, FILE **p_errfile,
                   SDL_Surface *surface, TTF_Font *font, int antialias,
                   int x, int y, int align, const char *text,
                   int red, int green, int blue, int alpha)
{
  int x0;
  int y0;
  SDL_Rect rect;
  SDL_Rect rect_src;
  SDL_Color font_color;
  SDL_Surface *rendered = NULL;
  SDL_Surface *temp_surface = NULL;
  SDL_Surface *temp_surface2 = NULL;

  /* sanity check */

  if (surface == NULL)
  {
    dbnj_fprintf(n_errfile, p_errfile,
                 "dbnj_draw_text: surface is NULL");
    return 1;
  }
  if (font == NULL)
  {
    dbnj_fprintf(n_errfile, p_errfile,
                 "dbnj_draw_text: font is NULL");
    return 1;
  }
  if (text == NULL)
  {
    dbnj_fprintf(n_errfile, p_errfile,
                 "dbnj_draw_text: text is NULL");
    return 1;
  }
  if ((red < 0) || (red > 255))
  {
    dbnj_fprintf(n_errfile, p_errfile,
                 "dbnj_draw_text: red (%d) is not in [0, 255]\n", red);
    return 1;
  }
  if ((green < 0) || (green > 255))
  {
    dbnj_fprintf(n_errfile, p_errfile,
                 "dbnj_draw_text: green (%d) is not in [0, 255]\n", green);
    return 1;
  }
  if ((blue < 0) || (blue > 255))
  {
    dbnj_fprintf(n_errfile, p_errfile,
                 "dbnj_draw_text: blue (%d) is not in [0, 255]\n", blue);
    return 1;
  }
  if ((alpha < 0) || (alpha > 255))
  {
    dbnj_fprintf(n_errfile, p_errfile,
                 "dbnj_draw_text: blue (%d) is not in [0, 255]\n", alpha);
    return 1;
  }

  font_color.r = red;
  font_color.g = green;
  font_color.b = blue;

  if (!antialias)
  {
    rendered = TTF_RenderUTF8_Solid(font, text, font_color);
    if (rendered == NULL)
    {
      dbnj_fprintf(n_errfile, p_errfile,
                   "dbnj_draw_text: TTF_RenderText_Solid failed");
      if (SDL_GetError() != NULL)
        dbnj_fprintf(n_errfile, p_errfile,
                     " (%s)", SDL_GetError());
      dbnj_fprintf(n_errfile, p_errfile,
                   "\n");
      return 1;
    }

    x0 = x;
    y0 = y;
    if (align == 1)
    {
      x0 = x - rendered->w;
    }
    else if (align == 2)
    {
      x0 = x - rendered->w / 2;
    }

    if (SDL_SetAlpha(rendered, SDL_SRCALPHA, alpha) != 0)
    {
      dbnj_fprintf(n_errfile, p_errfile,
                   "dbnj_draw_text: SDL_SetAlpha(rendered) failed");
      if (SDL_GetError() != NULL)
        dbnj_fprintf(n_errfile, p_errfile,
                     " (%s)", SDL_GetError());
      dbnj_fprintf(n_errfile, p_errfile,
                   "\n");
      SDL_FreeSurface(rendered);
      rendered = NULL;
      return 1;
    }

    rect.x = x0;
    rect.y = y0;
    if (SDL_BlitSurface(rendered, NULL, surface, &rect) != 0)
    {
      dbnj_fprintf(n_errfile, p_errfile,
                   "dbnj_draw_text: SDL_BlitSurface(rendered, surface) "
                   "failed");
      if (SDL_GetError() != NULL)
        dbnj_fprintf(n_errfile, p_errfile,
                     " (%s)", SDL_GetError());
      dbnj_fprintf(n_errfile, p_errfile,
                   "\n");
      SDL_FreeSurface(rendered);
      rendered = NULL;
      return 1;
    }

    SDL_FreeSurface(rendered);
    rendered = NULL;
    return 0;
  }
  else
  {
    rendered = TTF_RenderUTF8_Blended(font, text, font_color);
    if (rendered == NULL)
    {
      dbnj_fprintf(n_errfile, p_errfile,
                   "dbnj_draw_text: TTF_RenderText_Blended failed");
      if (SDL_GetError() != NULL)
        dbnj_fprintf(n_errfile, p_errfile,
                     " (%s)", SDL_GetError());
      dbnj_fprintf(n_errfile, p_errfile,
                   "\n");
      return 1;
    }

    x0 = x;
    y0 = y;
    if (align == 1)
    {
      x0 = x - rendered->w;
    }
    else if (align == 2)
    {
      x0 = x - rendered->w / 2;
    }

    temp_surface = SDL_CreateRGBSurface(SDL_SRCALPHA,
                                        rendered->w, rendered->h, 32,
                                        surface->format->Rmask,
                                        surface->format->Gmask, 
                                        surface->format->Bmask,
                                        0);
    if (temp_surface == NULL)
    {
      dbnj_fprintf(n_errfile, p_errfile,
                   "dbnj_draw_text: SDL_CreateRGBSurface failed");
      if (SDL_GetError() != NULL)
        dbnj_fprintf(n_errfile, p_errfile,
                     " (%s)", SDL_GetError());
      dbnj_fprintf(n_errfile, p_errfile,
                   "\n");
      SDL_FreeSurface(rendered);
      rendered = NULL;
      return 1;
    }

    /* SDL_DisplayFormat creates a new surface, so we must free temp_surface
     * before overwriting it 
     */
    temp_surface2 = SDL_DisplayFormat(temp_surface);
    if (temp_surface2 == NULL)
    {
      dbnj_fprintf(n_errfile, p_errfile,
                   "dbnj_draw_text: SDL_DisplayFormat failed");
      if (SDL_GetError() != NULL)
        dbnj_fprintf(n_errfile, p_errfile,
                     " (%s)", SDL_GetError());
      dbnj_fprintf(n_errfile, p_errfile,
                   "\n");
      SDL_FreeSurface(temp_surface);
      temp_surface = NULL;
      SDL_FreeSurface(rendered);
      rendered = NULL;
    }
    SDL_FreeSurface(temp_surface);
    temp_surface = temp_surface2;
    temp_surface2 = NULL;

    rect_src.x = x0;
    rect_src.y = y0;
    rect_src.w = rendered->w;
    rect_src.h = rendered->h;
    rect.x = 0;
    rect.y = 0;
    if (SDL_BlitSurface(surface, &rect_src, temp_surface, &rect) != 0)
    {
      dbnj_fprintf(n_errfile, p_errfile,
                   "dbnj_draw_text: SDL_BlitSurface(surface, temp_surface) "
                   "failed");
      if (SDL_GetError() != NULL)
        dbnj_fprintf(n_errfile, p_errfile,
                     " (%s)", SDL_GetError());
      dbnj_fprintf(n_errfile, p_errfile,
                   "\n");
      SDL_FreeSurface(temp_surface);
      temp_surface = NULL;
      SDL_FreeSurface(rendered);
      rendered = NULL;
      return 1;
    }

    rect.x = 0;
    rect.y = 0;
    if (SDL_BlitSurface(rendered, NULL, temp_surface, &rect) != 0)
    {
      dbnj_fprintf(n_errfile, p_errfile,
                   "dbnj_draw_text: SDL_BlitSurface(rendered, temp_surface) "
                   "failed");
      if (SDL_GetError() != NULL)
        dbnj_fprintf(n_errfile, p_errfile,
                     " (%s)", SDL_GetError());
      dbnj_fprintf(n_errfile, p_errfile,
                   "\n");
      SDL_FreeSurface(temp_surface);
      temp_surface = NULL;
      SDL_FreeSurface(rendered);
      rendered = NULL;
      return 1;
    }

    if (SDL_SetAlpha(temp_surface, SDL_SRCALPHA, alpha) != 0)
    {
      dbnj_fprintf(n_errfile, p_errfile,
                   "dbnj_draw_text: SDL_SetAlpha(temp_surface) failed");
      if (SDL_GetError() != NULL)
        dbnj_fprintf(n_errfile, p_errfile,
                     " (%s)", SDL_GetError());
      dbnj_fprintf(n_errfile, p_errfile,
                   "\n");
      SDL_FreeSurface(temp_surface);
      temp_surface = NULL;
      SDL_FreeSurface(rendered);
      rendered = NULL;
      return 1;
    }

    rect.x = x0;
    rect.y = y0;
    if (SDL_BlitSurface(temp_surface, NULL, surface, &rect) != 0)
    {
      dbnj_fprintf(n_errfile, p_errfile,
                   "dbnj_draw_text: SDL_BlitSurface(temp_surface, surface) "
                   "failed");
      if (SDL_GetError() != NULL)
        dbnj_fprintf(n_errfile, p_errfile,
                     " (%s)", SDL_GetError());
      dbnj_fprintf(n_errfile, p_errfile,
                   "\n");
      SDL_FreeSurface(temp_surface);
      temp_surface = NULL;
      SDL_FreeSurface(rendered);
      rendered = NULL;
      return 1;
    }

    SDL_FreeSurface(temp_surface);
    temp_surface = NULL;
    SDL_FreeSurface(rendered);
    rendered = NULL;
    return 0;
  }

  return 0;
}
