/* $Id: dbnj_util.c,v 1.30 2005/11/02 11:55:43 oohara Exp $ */

/* vfprintf */
#include <stdio.h>
/* strtol, malloc */
#include <stdlib.h>
/* errno */
#include <errno.h>
/* INT_MIN, INT_MAX */
#include <limits.h>
/* strlen, strcpy, strcmp */
#include <string.h>
#include <stdarg.h>

#include "dbnj_util.h"

/* formatted output to each file in the same way as fprintf does
 * p_file (arg 2) must be exactly n_file (arg 1) pointers to FILE
 * return 0 on success, 1 on error
 */
int
dbnj_fprintf(int n_file, FILE **p_file, const char *template, ...)
{
  va_list ap;
  int i;
  int status = 0;

  /* sanity check */
  if (n_file < 0)
    return 1;
  if (n_file == 0)
  {
    /* this is not an error */
    return 0;
  }
  if (p_file == NULL)
    return 1;
  if (template == NULL)
    return 1;

  for (i = 0; i < n_file; i++)
  {
    if (p_file[i] == NULL)
    {
      status = 1;
      break;
    }

    va_start(ap, template);
    if (vfprintf(p_file[i], template, ap) == EOF)
      status = 1;
    va_end(ap);
    if (fflush(p_file[i]) == EOF)
      status = 1;

    if (status != 0)
      break;
  }

  return status;
}

/* convert string (arg 4) to a number and store it to value (arg 3)
 * return 0 on success, 1 on error
 */
int dbnj_str_to_int(int n_errfile, FILE **p_errfile,
                    int *value, const char *string)
{
  long int temp;
  char *tail;

  /* sanity check */
  if (value == NULL)
  {
    dbnj_fprintf(n_errfile, p_errfile,
                 "dbnj_str_to_int: value is NULL\n");
    return 1;
  }
  if (string == NULL)
  {
    dbnj_fprintf(n_errfile, p_errfile,
                 "dbnj_str_to_int: string is NULL\n");
    return 1;
  }

  /* to detect overflow */
  errno = 0;
  temp = strtol(string, &tail, 10);
  if (tail[0] != '\0')
  {
    dbnj_fprintf(n_errfile, p_errfile,
                 "dbnj_str_to_int: string is not a number\n");
    return 1;
  }
  if (errno != 0)
  {
    dbnj_fprintf(n_errfile, p_errfile,
                 "dbnj_str_to_int: number overflowed\n");
    return 1;
  }
  if ((temp <= INT_MIN) || (temp >= INT_MAX))
  {
    dbnj_fprintf(n_errfile, p_errfile,
                 "dbnj_str_to_int: too big or small integer\n");
    return 1;
  }
  *value = (int) temp;

  return 0;
}

/* add ... (arg 5 --) at the end of result (arg 3)
 * ... must be exactly n (arg 4) pointers to const char
 * if result (arg 3) is not NULL, its memory must be allocated
 * with malloc(); it is resized as necessary
 * ... (arg 5 --) are copied
 * return NULL on error
 */
char *
dbnj_concat_string(int n_errfile, FILE **p_errfile,
                   char *result, int n, ...)
{
  va_list ap;
  int i;
  int length;
  int head;
  const char *s = NULL;
  char *temp = NULL;

  /* sanity check */
  if (n < 0)
  {
    dbnj_fprintf(n_errfile, p_errfile,
                 "dbnj_concat_string: n is negative (%d)\n", n);
    return NULL;
  }

  if (n == 0)
    return result;

  if (result == NULL)
    head = 0;
  else
    head = (int) strlen(result);
  length = head;

  va_start(ap, n);
  for (i = 0; i < n; i++)
  {
    s = va_arg(ap, const char *);
    if (s == NULL)
    {
      dbnj_fprintf(n_errfile, p_errfile,
                   "dbnj_concat_string: string %d is NULL\n", i + 1);
      return NULL;
    }
    length += (int) strlen(s);
  }
  va_end(ap);

  if (length < 0)
  {
    dbnj_fprintf(n_errfile, p_errfile,
                 "dbnj_concat_string: length is negative (%d)\n", length);
    return NULL;
  }
  if (length == 0)
    return result;

  /* +1 is for the trailing '\0' */
  if (result == NULL)
  {
    temp = (char *) malloc(sizeof(char) * (length + 1));
  }
  else
  {
    temp = (char *) realloc(result, sizeof(char) * (length + 1));
  }
  if (temp == NULL)
  {
    dbnj_fprintf(n_errfile, p_errfile,
                 "dbnj_concat_string: can't allocate enough memory\n");
    return NULL;
  }
  result = temp;

  va_start(ap, n);
  for (i = 0; i < n; i++)
  {
    s = va_arg(ap, const char *);
    strcpy(result + head, s);
    head += (int) strlen(s);
  }
  va_end(ap);

  result[length] = '\0';
  return result;
}

int
dbnj_string_hash(const char *s)
{
  int c;
  int h;
  int h_high;
  int h_low;

  h = 0;
  c = *s;
  while (c != '\0')
  {
    h_high = h & 0x7c000000;
    h_low =  h & 0x03ffffff;
    h = (h_high >> 26) | (h_low << 5);
    h = h ^ c;

    s++;
    c = *s;
  }

  if (h < 0)
    h = -h;

  return h;
}

int
dbnj_string_equal(const char *s1, const char *s2)
{
  return !strcmp(s1, s2);
}

/* rgb (arg 1) must be int[3] (you must be enough memory before calling
 * this function)
 * return 0 on success, 1 on error
 */
int dbnj_hsv_to_rgb(int n_errfile, FILE **p_errfile,
                    int *rgb, int hue, int saturation, int value)
{
  int red;
  int green;
  int blue;

  /* sanity check */
  if (rgb == NULL)
  {
    dbnj_fprintf(n_errfile, p_errfile,
                 "dbnj_hsv_to_rgb: rgb is NULL\n");
    return 1;
  }
  if ((hue < 0) || (hue > 360))
  {
    dbnj_fprintf(n_errfile, p_errfile,
                 "dbnj_hsv_to_rgb: hue (%d) is not in [0, 360]\n", hue);
    return 1;
  }
  if ((saturation < 0) || (saturation > 100))
  {
    dbnj_fprintf(n_errfile, p_errfile,
                 "dbnj_hsv_to_rgb: saturation (%d) is not in [0, 100]\n",
                 saturation);
    return 1;
  }
  if ((value < 0) || (value > 100))
  {
    dbnj_fprintf(n_errfile, p_errfile,
                 "dbnj_hsv_to_rgb: value (%d) is not in [0, 100]\n", value);
    return 1;
  }

  red = 0;
  green = 0;
  blue = 0;
  if (hue < 60)
  {
    red = 255;
    green = 255 * hue / 60;
    blue = 0;
  }
  else if (hue < 120)
  {
    red = 255 * (120 - hue) / 60;
    green = 255;
    blue = 0;
  }
  else if (hue < 180)
  {
    red = 0;
    green = 255;
    blue = 255 * (hue - 120) / 60;
  }
  else if (hue < 240)
  {
    red = 0;
    green = 255 * (240 - hue) / 60;
    blue = 255;
  }
  else if (hue < 300)
  {
    red = 255 * (hue - 240) / 60;
    green = 0;
    blue = 255;
  }
  else
  {
    red = 255;
    green = 0;
    blue = 255 * (360 - hue) / 60;
  }

  if (red < 0)
    red = 0;
  if (red > 255)
    red = 255;
  if (green < 0)
    green = 0;
  if (green > 255)
    green = 255;
  if (blue < 0)
    blue = 0;
  if (blue > 255)
    blue = 255;

  red = (red * saturation + 255 * (100 - saturation)) / 100;
  green = (green * saturation + 255 * (100 - saturation)) / 100;
  blue = (blue * saturation + 255 * (100 - saturation)) / 100;

  if (red < 0)
    red = 0;
  if (red > 255)
    red = 255;
  if (green < 0)
    green = 0;
  if (green > 255)
    green = 255;
  if (blue < 0)
    blue = 0;
  if (blue > 255)
    blue = 255;

  red = red * value / 100;
  green = green * value / 100;
  blue = blue * value / 100;

  if (red < 0)
    red = 0;
  if (red > 255)
    red = 255;
  if (green < 0)
    green = 0;
  if (green > 255)
    green = 255;
  if (blue < 0)
    blue = 0;
  if (blue > 255)
    blue = 255;

  rgb[0] = red;
  rgb[1] = green;
  rgb[2] = blue;

  return 0;
}
