/* $Id: markov.c,v 1.36 2005/10/26 13:40:41 oohara Exp $ */

#include <stdio.h>
/* rand, malloc */
#include <stdlib.h>

#include "dbnj_util.h"
#include "dbnj_hash_table.h"
#include "markov_lexical.h"

#include "markov.h"

/* return NULL on error */
markov *
markov_new(int n_errfile, FILE **p_errfile)
{
  markov *result = NULL;

  result = (markov *) malloc(sizeof(markov));
  if (result == NULL)
  {
    dbnj_fprintf(n_errfile, p_errfile,
                 "markov_new: malloc(result) failed\n");
    return NULL;
  }

  result->hash_table = NULL;
  result->n_state = 0;
  result->p_state = NULL;
  result->n_next = NULL;
  result->p_next = NULL;
  result->previous = NULL;

  result->hash_table = dbnj_hash_table_new(n_errfile, p_errfile,
                                           (int (*)(const void *)) 
                                           dbnj_string_hash,
                                           (int (*)(const void *,
                                                    const void *))
                                           dbnj_string_equal,
                                           (void (*)(void *)) &free);
  if (result->hash_table == NULL)
  {
    dbnj_fprintf(n_errfile, p_errfile,
                 "markov_new: dbnj_hash_table_new failed\n");
    markov_delete(n_errfile, p_errfile,
                result);
    return NULL;
  }
  
  return result;
}

void
markov_delete(int n_errfile, FILE **p_errfile,
              markov *chain)
{
  int i;
  int j;
  char **p;

  if (chain == NULL)
    return;

  if (chain->hash_table != NULL)
  {
    dbnj_hash_table_delete(n_errfile, p_errfile,
                           chain->hash_table);
    chain->hash_table = NULL;
  }

  if (chain->p_state != NULL)
  {
    for (i = 0; i < chain->n_state; i++)
    {
      if (chain->p_state[i] != NULL)
        free(chain->p_state[i]);
      chain->p_state[i] = NULL;
    }
    free(chain->p_state);
    chain->p_state = NULL;
  }

  if (chain->p_next != NULL)
  {
    for (i = 0; i < chain->n_state; i++)
    {
      p = chain->p_next[i];
      for (j = 0; j < chain->n_next[i]; j++)
      {
        free(p[j]);
        p[j] = NULL;
      }
      free(chain->p_next[i]);
      chain->p_next[i] = NULL;
    }
    chain->p_next = NULL;
  }

  /* don't free chain->previous */

  chain->hash_table = NULL;
  chain->n_state = 0;
  chain->p_state = NULL;
  chain->n_next = NULL;
  chain->p_next = NULL;
  chain->previous = NULL;

  free(chain);
}

/* state (arg 4) and next (arg 5) are copied
 * return 0 on success, 1 on error
 */
int
markov_insert(int n_errfile, FILE **p_errfile,
              markov *chain,
              const char *state, const char *next)
{
  int n;
  int found;
  char *key = NULL;
  int *data = NULL;
  char *state_copied = NULL;
  char *next_copied = NULL;
  char **p = NULL;
  char **p_state_new = NULL;
  int *n_next_new = NULL;
  char ***p_next_new_0 = NULL;
  char **p_next_new_1 = NULL;
  void *found_data = NULL;

  /* sanity check */
  if (chain == NULL)
  {
    dbnj_fprintf(n_errfile, p_errfile,
                 "markov_insert: chain is NULL\n");
    return 1;
  }
  if (state == NULL)
  {
    dbnj_fprintf(n_errfile, p_errfile,
                 "markov_insert: state is NULL\n");
    return 1;
  }
  if (next == NULL)
  {
    dbnj_fprintf(n_errfile, p_errfile,
                 "markov_insert: next is NULL\n");
    return 1;
  }

  found_data = NULL;
  found = dbnj_hash_table_lookup(n_errfile, p_errfile,
                                 chain->hash_table,
                                 (const void *) state,
                                 NULL, &found_data);
  if (!found)
  {
    key = dbnj_concat_string(n_errfile, p_errfile,
                             NULL, 1, state);
    if (key == NULL)
    {
      dbnj_fprintf(n_errfile, p_errfile,
                   "markov_insert: dbnj_concat_string(key) failed\n");
      return 1;
    }

    data = (int *) malloc(sizeof(int));
    if (data == NULL)
    {
      dbnj_fprintf(n_errfile, p_errfile,
                   "markov_insert: malloc(data) failed\n");
      free(key);
      return 1;
    }
    n = chain->n_state;
    *data = n;

    state_copied = dbnj_concat_string(n_errfile, p_errfile,
                                      NULL, 1, state);
    if (state_copied == NULL)
    {
      dbnj_fprintf(n_errfile, p_errfile,
                 "markov_insert: dbnj_concat_string(state_copied) failed\n");
      free(data);
      free(key);
      return 1;
    }

    if (chain->p_state == NULL)
      p_state_new = (char **) malloc(sizeof(char *) * 1);
    else
      p_state_new = (char **) realloc(chain->p_state,
                                      sizeof(char *) * (chain->n_state + 1));
    if (p_state_new == NULL)
    {
      dbnj_fprintf(n_errfile, p_errfile,
                 "markov_insert: malloc(p_state_new) failed\n");
      free(state_copied);
      free(data);
      free(key);
      return 1;
    }
    p_state_new[chain->n_state] = NULL;

    if (chain->n_next == NULL)
      n_next_new = (int *) malloc(sizeof(int) * 1);
    else
      n_next_new = (int *) realloc(chain->n_next,
                                   sizeof(int) * (chain->n_state + 1));
    if (n_next_new == NULL)
    {
      dbnj_fprintf(n_errfile, p_errfile,
                 "markov_insert: malloc(n_next_new) failed\n");
      if (chain->n_next == NULL)
        free(n_next_new);
      free(state_copied);
      free(data);
      free(key);
      return 1;
    }
    n_next_new[chain->n_state] = 0;

    if (chain->p_next == NULL)
      p_next_new_0 = (char ***) malloc(sizeof(char **) * 1);
    else
      p_next_new_0 = (char ***) realloc(chain->p_next,
                                        sizeof(char **) * (chain->n_state +1));
    if (p_next_new_0 == NULL)
    {
      dbnj_fprintf(n_errfile, p_errfile,
                 "markov_insert: malloc(p_next_new_0) failed\n");
      if (chain->p_next == NULL)
        free(p_next_new_0);
      if (chain->n_next == NULL)
        free(n_next_new);
      free(state_copied);
      free(data);
      free(key);
      return 1;
    }
    p_next_new_0[chain->n_state] = NULL;
      
    if (dbnj_hash_table_insert(n_errfile, p_errfile,
                               chain->hash_table, 0,
                               (void *) key, (void *) data,
                               (void (*)(void *)) &free) != 0)
    {
      dbnj_fprintf(n_errfile, p_errfile,
                   "markov_insert: dbnj_hash_table_insert failed\n");
      free(state_copied);
      free(data);
      free(key);
      return 1;
    }

    chain->p_state = p_state_new;
    chain->n_next = n_next_new;
    chain->p_next = p_next_new_0;
    chain->p_state[n] = state_copied;
    (chain->n_state)++;
  }
  else
  {
    n = *((int *) found_data);
  }

  next_copied = dbnj_concat_string(n_errfile, p_errfile,
                                   NULL, 1, next);
  if (next_copied == NULL)
  {
    dbnj_fprintf(n_errfile, p_errfile,
                 "markov_insert: dbnj_concat_string(next_copied) failed\n");
    return 1;
  }

  if (chain->p_next[n] == NULL)
    p_next_new_1 = (char **) malloc(sizeof(char *) * 1);
  else
    p_next_new_1 = (char **) realloc(chain->p_next[n],
                                     sizeof(char *) * (chain->n_next[n] + 1));
  if (p_next_new_1 == NULL)
  {
    dbnj_fprintf(n_errfile, p_errfile,
                 "markov_insert: malloc(p_next_new_1) failed\n");
    free(next_copied);
    return 1;
  }
  p_next_new_1[chain->n_next[n]] = NULL;

  chain->p_next[n] = p_next_new_1;
  p = chain->p_next[n];
  p[chain->n_next[n]] = next_copied;
  (chain->n_next[n])++;

  return 0;
}

/* return NULL on error */
const char *
markov_next(int n_errfile, FILE **p_errfile,
            markov *chain)
{
  int decided;
  int found;
  int n;
  int a;
  void *found_data = NULL;
  char **p = NULL;
  const char *result = NULL;

  /* sanity check */
  if (chain == NULL)
  {
    dbnj_fprintf(n_errfile, p_errfile,
                 "markov_next: chain is NULL\n");
    return NULL;
  }
  if (chain->n_state <= 0)
  {
    dbnj_fprintf(n_errfile, p_errfile,
                 "markov_next: chain->n_state is non-positive (%d)\n",
                 chain->n_state);
    return NULL;
  }
  if (chain->p_state == NULL)
  {
    dbnj_fprintf(n_errfile, p_errfile,
                 "markov_next: chain->p_state is NULL\n");
    return NULL;
  }
  if (chain->n_next == NULL)
  {
    dbnj_fprintf(n_errfile, p_errfile,
                 "markov_next: chain->n_next is NULL\n");
    return NULL;
  }
  if (chain->p_next == NULL)
  {
    dbnj_fprintf(n_errfile, p_errfile,
                 "markov_next: chain->p_next is NULL\n");
    return NULL;
  }

  decided = 0;
  if (chain->previous != NULL)
  {  
    found_data = NULL;
    found = dbnj_hash_table_lookup(n_errfile, p_errfile,
                                   chain->hash_table,
                                   (const void *) chain->previous,
                                   NULL, &found_data);
    if (found)
    {
      n = *((int *) found_data);
      p = chain->p_next[n];
      if ((p != NULL) && (chain->n_next[n] > 0))
      {
        a = rand() % (chain->n_next[n]);
        result = (const char *) (p[a]);
        if (!dbnj_string_equal(result, chain->previous))
          decided = 1;
      }
    }
  }

  if (!decided)
  {
    a = rand() % (chain->n_state);
    result = (const char *) (chain->p_state[a]);
  }

  if (result == NULL)
  {
    dbnj_fprintf(n_errfile, p_errfile,
                 "markov_next: NULL was chosen\n");
    return NULL;
  }

  chain->previous = result;
  return result;
}
